@@ -1,17 +1,23 @@
 module Network.TLS.Parameters (
     ClientParams (..),
+    defaultParamsClient,
     ServerParams (..),
+    defaultParamsServer,
     CommonParams,
     DebugParams (..),
+    defaultDebugParams,
     ClientHooks (..),
+    defaultClientHooks,
     OnCertificateRequest,
     OnServerCertificate,
     ServerHooks (..),
+    defaultServerHooks,
     Supported (..),
+    defaultSupported,
     Shared (..),
-
-    -- * special default
-    defaultParamsClient,
+    defaultShared,
+    Limit (..),
+    defaultLimit,
 
     -- * Parameters
     MaxFragmentEnum (..),
@@ -19,15 +25,18 @@ module Network.TLS.Parameters (
     GroupUsage (..),
     CertificateUsage (..),
     CertificateRejectReason (..),
+    Information (..),
 ) where
 
 import qualified Data.ByteString as B
-import Data.Default.Class
+import Data.Default (Default (def))
 import Network.TLS.Cipher
 import Network.TLS.Compression
 import Network.TLS.Credentials
 import Network.TLS.Crypto
 import Network.TLS.Extension
+import Network.TLS.Extra.Cipher
+import Network.TLS.Handshake.State
 import Network.TLS.Imports
 import Network.TLS.Measurement
 import Network.TLS.RNG (Seed)
@@ -37,7 +46,7 @@ import qualified Network.TLS.Struct as Struct
 import Network.TLS.Types (HostName)
 import Network.TLS.X509
 
-type CommonParams = (Supported, Shared, DebugParams)
+type CommonParams = (Supported, Shared, DebugParams, Limit)
 
 -- | All settings should not be used in production
 data DebugParams = DebugParams
@@ -62,6 +71,7 @@ data DebugParams = DebugParams
     -- Default: no printing
     }
 
+-- | Default value for 'DebugParams'
 defaultDebugParams :: DebugParams
 defaultDebugParams =
     DebugParams
@@ -76,6 +86,8 @@ instance Show DebugParams where
 instance Default DebugParams where
     def = defaultDebugParams
 
+{-# DEPRECATED clientUseMaxFragmentLength "UseMaxFragmentLength is deprecated" #-}
+
 data ClientParams = ClientParams
     { clientUseMaxFragmentLength :: Maybe MaxFragmentEnum
     -- ^
@@ -99,9 +111,19 @@ data ClientParams = ClientParams
     --
     -- Default: 'True'
     , clientWantSessionResume :: Maybe (SessionID, SessionData)
-    -- ^ try to establish a connection using this session.
+    -- ^ try to establish a connection using this session for TLS 1.2/TLS 1.3.
+    -- This can be used for TLS 1.3 but for backward compatibility purpose only.
+    -- Use 'clientWantSessionResume13' instead for TLS 1.3.
     --
     -- Default: 'Nothing'
+    , clientWantSessionResumeList :: [(SessionID, SessionData)]
+    -- ^ try to establish a connection using one of this sessions
+    -- especially for TLS 1.3.
+    -- This take precedence over 'clientWantSessionResume'.
+    -- For convenience, this can be specified for TLS 1.2 but only the first
+    -- entry is used.
+    --
+    -- Default: '[]'
     , clientShared :: Shared
     -- ^ See the default value of 'Shared'.
     , clientHooks :: ClientHooks
@@ -120,9 +142,11 @@ data ClientParams = ClientParams
     -- is automatically re-sent.
     --
     -- Default: 'False'
+    , clientLimit :: Limit
     }
     deriving (Show)
 
+-- | Default value for 'ClientParams'
 defaultParamsClient :: HostName -> ByteString -> ClientParams
 defaultParamsClient serverName serverId =
     ClientParams
@@ -130,11 +154,13 @@ defaultParamsClient serverName serverId =
         , clientServerIdentification = (serverName, serverId)
         , clientUseServerNameIndication = True
         , clientWantSessionResume = Nothing
+        , clientWantSessionResumeList = []
         , clientShared = def
         , clientHooks = def
         , clientSupported = def
         , clientDebug = defaultDebugParams
         , clientUseEarlyData = False
+        , clientLimit = defaultLimit
         }
 
 data ServerParams = ServerParams
@@ -176,6 +202,7 @@ data ServerParams = ServerParams
     -- Acceptable value range is 0 to 604800 (7 days).
     --
     -- Default: 7200 (2 hours)
+    , serverLimit :: Limit
     }
     deriving (Show)
 
@@ -191,6 +218,7 @@ defaultParamsServer =
         , serverDebug = defaultDebugParams
         , serverEarlyDataSize = 0
         , serverTicketLifetime = 7200
+        , serverLimit = defaultLimit
         }
 
 instance Default ServerParams where
@@ -313,7 +341,7 @@ data Supported = Supported
     --   The default value includes all groups with security strength of 128
     --   bits or more.
     --
-    --   Default: @[X25519,X448,P256,FFDHE3072,FFDHE4096,P384,FFDHE6144,FFDHE8192,P521]@
+    --   Default: @[X25519,X448,P256,FFDHE2048,FFDHE3072,FFDHE4096,P384,FFDHE6144,FFDHE8192,P521]@
     }
     deriving (Show, Eq)
 
@@ -331,7 +359,7 @@ defaultSupported :: Supported
 defaultSupported =
     Supported
         { supportedVersions = [TLS13, TLS12]
-        , supportedCiphers = []
+        , supportedCiphers = ciphersuite_default
         , supportedCompressions = [nullCompression]
         , supportedHashSignatures = Struct.supportedSignatureSchemes
         , supportedSecureRenegotiation = True
@@ -395,14 +423,17 @@ data Shared = Shared
 instance Show Shared where
     show _ = "Shared"
 instance Default Shared where
-    def =
-        Shared
-            { sharedCredentials = mempty
-            , sharedSessionManager = noSessionManager
-            , sharedCAStore = mempty
-            , sharedValidationCache = def
-            , sharedHelloExtensions = []
-            }
+    def = defaultShared
+
+defaultShared :: Shared
+defaultShared =
+    Shared
+        { sharedCredentials = mempty
+        , sharedSessionManager = noSessionManager
+        , sharedCAStore = mempty
+        , sharedValidationCache = def
+        , sharedHelloExtensions = []
+        }
 
 -- | Group usage callback possible return values.
 data GroupUsage
@@ -534,6 +565,8 @@ data ClientHooks = ClientHooks
     --   (4) rejecting if dh_size < 1024 (to prevent Logjam attack)
     --
     --   See RFC 7919 section 3.1 for recommandations.
+    , onServerFinished :: Information -> IO ()
+    -- ^ When a handshake is done, this hook can check `Information`.
     }
 
 defaultClientHooks :: ClientHooks
@@ -543,6 +576,7 @@ defaultClientHooks =
         , onServerCertificate = validateDefault
         , onSuggestALPN = return Nothing
         , onCustomFFDHEGroup = defaultGroupUsage 1024
+        , onServerFinished = \_ -> return ()
         }
 
 instance Show ClientHooks where
@@ -614,6 +648,7 @@ data ServerHooks = ServerHooks
     -- Default: 'return'
     }
 
+-- | Default value for 'ServerHooks'
 defaultServerHooks :: ServerHooks
 defaultServerHooks =
     ServerHooks
@@ -622,7 +657,9 @@ defaultServerHooks =
                 CertificateUsageReject $
                     CertificateRejectOther "no client certificates expected"
         , onUnverifiedClientCert = return False
-        , onCipherChoosing = \_ -> head
+        , onCipherChoosing = \_ ccs -> case ccs of
+            [] -> error "onCipherChoosing"
+            c : _ -> c
         , onServerNameIndication = \_ -> return mempty
         , onNewHandshake = \_ -> return True
         , onALPNClientSuggest = Nothing
@@ -633,3 +670,52 @@ instance Show ServerHooks where
     show _ = "ServerHooks"
 instance Default ServerHooks where
     def = defaultServerHooks
+
+-- | Information related to a running context, e.g. current cipher
+data Information = Information
+    { infoVersion :: Version
+    , infoCipher :: Cipher
+    , infoCompression :: Compression
+    , infoMainSecret :: Maybe ByteString
+    , infoExtendedMainSecret :: Bool
+    , infoClientRandom :: Maybe ClientRandom
+    , infoServerRandom :: Maybe ServerRandom
+    , infoSupportedGroup :: Maybe Group
+    , infoTLS12Resumption :: Bool
+    , infoTLS13HandshakeMode :: Maybe HandshakeMode13
+    , infoIsEarlyDataAccepted :: Bool
+    }
+    deriving (Show, Eq)
+
+-- | Limitations for security.
+--
+-- @since 2.1.7
+data Limit = Limit
+    { limitRecordSize :: Maybe Int
+    -- ^ Record size limit defined in RFC 8449.
+    --
+    -- If 'Nothing', the "record_size_limit" extension is not used.
+    --
+    -- In the case of 'Just': A client sends the "record_size_limit"
+    -- extension with this value to the server. A server sends back
+    -- this extension with its own value if a client sends the
+    -- extension. When negotiated, both my limit and peer's limit
+    -- are enabled for protected communication.
+    --
+    -- Default: Nothing
+    , limitHandshakeFragment :: Int
+    -- ^ The limit to accept the number of each handshake message.
+    -- For instance, a nasty client may send many fragments of client
+    -- certificate.
+    --
+    -- Default: 32
+    }
+    deriving (Eq, Show)
+
+-- | Default value for 'Limit'.
+defaultLimit :: Limit
+defaultLimit =
+    Limit
+        { limitRecordSize = Nothing
+        , limitHandshakeFragment = 32
+        }
